/*
 * earthapp.cpp
 *
 * $Id: earthapp.cpp,v 0.1 1998/12/10 20:07:16 thorsten Exp thorsten $
 *
 * This file is part of XGlobe. See README for details.
 *
 * Copyright (C) 1998 Thorsten Scheuermann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as published by
 * the Free Software Foundation.
 *
 * $Log: earthapp.cpp,v $
 * Revision 0.1  1998/12/10 20:07:16  thorsten
 * initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "config.h"
#include "earthapp.h"

/* ------------------------------------------------------------------------*/

EarthApplication::EarthApplication(int &argc, char **argv)
    : QApplication(argc, argv)
{
  firstTime = TRUE;
  view_lat = 0.;
  view_long = 0.;
  delay = 300;
  zoom = 0.75;
  p_type = SUNREL;
  builtin_markers = TRUE;
  show_markers = TRUE;
  show_label = TRUE;
  label_x = -5;
  label_y = 5;
  ambient_light = 0.15;
  fov = -1.;
  with_nightmap = FALSE;
  argc_map = -1;
  argc_nightmap = -1;
  do_the_dump = FALSE;
  dwidget = NULL;
  use_kde = FALSE;
  
  // evaluate command line parameters
  for(int i=1; i<this->argc(); i++)
  {
    if(strcmp(this->argv()[i], "-pos") == 0)
    {
      i = readPosition(i+1);
    }
    else if(strcmp(this->argv()[i], "-wait") == 0)
    {
      readDelay(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-zoom") == 0)
    {
      readZoom(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-nice") == 0)
    {
      readPriority(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-markers") == 0)
    {
      builtin_markers = TRUE;
      show_markers = TRUE;
    }
    else if(strcmp(this->argv()[i], "-nomarkers") == 0)
    {
      builtin_markers = FALSE;
      show_markers = FALSE;
    }
    else if(strcmp(this->argv()[i], "-markerfile") == 0)
    {
      readMarkerFile(i+1);
      show_markers = TRUE;
      i++;
    }
    else if(strcmp(this->argv()[i], "-label") == 0)
    {
      show_label = TRUE;
    }
    else if(strcmp(this->argv()[i], "-nolabel") == 0)
    {
      show_label = FALSE;
    }
    else if(strcmp(this->argv()[i], "-labelpos") == 0)
    {
      readLabelPos(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-ambientlight") == 0)
    {
      readAmbientLight(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-fov") == 0)
    {
      readFov(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-nightmap") == 0)
    {
      with_nightmap = TRUE;
    }
    else if(strcmp(this->argv()[i], "-nonightmap") == 0)
    {
      with_nightmap = FALSE;
    }
    else if(strcmp(this->argv()[i], "-mapfile") == 0)
    {
      readMapFile(i+1);
      i++;
    }
    else if(strcmp(this->argv()[i], "-nightmapfile") == 0)
    {
      readNightMapFile(i+1);
      i++;
      with_nightmap = TRUE;
    }
    else if(strcmp(this->argv()[i], "-help") == 0)
    {
      printHelp();
      ::exit(0);
    }
    else if(strcmp(this->argv()[i], "-dump") == 0)
    {
      do_the_dump = TRUE;
    }
    else if(strcmp(this->argv()[i], "-kde") == 0)
    {
      use_kde = TRUE;
    }
    else
    {
      printUsage();
      ::exit(1);
    }
  }

  if(use_kde && !do_the_dump)
  {
    dwidget = new DesktopWidget();
    ASSERT(dwidget != NULL);
    dwidget->update();
  }
}

/* ------------------------------------------------------------------------*/

EarthApplication::~EarthApplication(void)
{
  ASSERT(r != NULL);
  delete r;
  ASSERT(timer != NULL);
  timer->stop();
  delete timer;

  if(dwidget)
    delete dwidget;
}

/* ------------------------------------------------------------------------*/

int EarthApplication::readPosition(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  if(strcmp(argv()[i], "fixed") == 0)
  {
    i++;
    if(i+1 >= argc())
    {
      printUsage();
      ::exit(1);
    }
    p_type = FIXED;
    view_lat = atof(argv()[i++]);
    view_long = atof(argv()[i]);
  }
  else if(strcmp(argv()[i], "sunrel") == 0)
  {
    i++;
    if(i+1 >= argc())
    {
      printUsage();
      ::exit(1);
    }
    p_type = SUNREL;
    view_lat = atof(argv()[i++]);
    view_long = atof(argv()[i]);
  }
  else if(strcmp(argv()[i], "random") == 0)
  {
    p_type = RANDOM;
    srand((int)time(NULL) + (int)getpid());
  }
  else
  {
    printUsage();
    ::exit(1);
  }
  return i;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readDelay(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  delay = atoi(argv()[i]);
  if(delay < 1)
    delay = 300;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readZoom(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  zoom = atof(argv()[i]);
  if(zoom <= 0.)
    zoom = 75.;
  zoom /= 100.;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readPriority(int i)
{
  int pri;
  
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  pri = atoi(argv()[i]);
  setPriority(pri);
}

/* ------------------------------------------------------------------------*/


void EarthApplication::readMarkerFile(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  if(!marker_list.loadMarkerFile(argv()[i]))
  {
    // bail out
    ::exit(1);
  }
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readLabelPos(int i)
{
  char *s;
  int  n;
  
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }

  s = argv()[i];

  switch(*s)
  {
    case '+':
      label_x = 1;
      break;
      
    case '-':
      label_x = -1;
      break;

    default:
    printUsage();
    ::exit(1);
  }
  
  n = atoi(++s);
  if(n != 0)             // to preserve the sign when 0 is used as parameter
    label_x *= n;
  
  while(isdigit(*++s)) ;

  switch(*s)
  {
    case '+':
      label_y = 1;
      break;
      
    case '-':
      label_y = -1;
      break;

    default:
    printUsage();
    ::exit(1);
  }
  n = atoi(++s);
  if(n != 0)             // to preserve the sign when 0 is used as parameter
    label_y *= n;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readAmbientLight(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  ambient_light = atof(argv()[i]);
  if(ambient_light > 100.)
    ambient_light = 1.;
  else if(ambient_light < 0.)
    ambient_light = 0.;

  ambient_light /= 100.;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readFov(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  fov = atof(argv()[i]);
  if(fov <= 0)
    fov = -1.;
  else if(fov >= 90.)
    fov = -1.;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readMapFile(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  argc_map = i;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::readNightMapFile(int i)
{
  if(i >= argc())
  {
    printUsage();
    ::exit(1);
  }
  argc_nightmap = i;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::randomPosition(void)
{
  view_lat = ((rand()%30001)/30000.)*180. - 90.;
  view_long = ((rand()%30001)/30000.)*360. - 180.;
}

/* ------------------------------------------------------------------------*/

void EarthApplication::setPriority(int pri)
{
  setpriority(PRIO_PROCESS, getpid(), pri);
}

/* ------------------------------------------------------------------------*/

void EarthApplication::printUsage(void)
{
  printf("XGlobe %s\n", VERSION);
  printf("Usage: %s [-pos pos_spec ] [-wait seconds ] [-zoom zoom]\n"
         "[-markers|-nomarkers] [-markerfile file] [-label|-nolabel] [-labelpos geom]\n"
         "[-ambientlight level] [-nice priority] [-nightmap|-nonightmap]\n"
         "[-mapfile file] [-nightmapfile file] [-kde] [-dump] [-help]\n"
         "For an explanation of command line options, use the switch -help.\n",
         argv()[0]);
}

/* ------------------------------------------------------------------------*/

void EarthApplication::printHelp(void)
{
  printf("XGlobe %s command line options:\n\n", VERSION);
  printf("-pos pos_spec  pos_spec consists of one of the keywords \"fixed\", \"sunrel\"\n"
         "               or \"random\", possibly followed by additional arguments:\n"
         "               fixed: Followed by two numerical arguments latitude and\n"
         "               longitude (given in decimal degrees) of a viewing position.\n"
         "                 positive lat. - north of equator\n"
         "                 negative lat. - south of equator\n"
         "                 positive long. - west of Greenwich\n"
         "                 negative long. - east of Greenwich\n"
         "               sunrel: Followed by to numerical arguments lat. and long.,\n"
         "               indicating offsets of current sun position.\n"
         "               random: selects a random viewing position each time a frame is\n"
         "               redrawn.\n\n");
  printf("-wait secs     Specifies the interval in seconds between screen updates.\n\n");
  printf("-zoom zoom     Specifies the size of the globe in relation to the screen size.\n"
         "               100 means diameter of globe equals screen height.\n\n");
  printf("-markers       Enable displaying of built-in display markers.\n"
         "-nomarkers     Disable displaying of built-in display markers.\n\n");
  printf("-markerfile file  Load an additional location marker file. (Have a look at\n"
         "                  file \"xglobe-markers\" for reference.)\n\n");
  printf("-label         Enable displaying of label indicating current date and time etc.\n"
         "-nolabel       Disable displaying of label\n\n");
  printf("-labelpos geom  geom specifies the screen location of the label.\n"
         "                Syntax: {+-}<xoffset>{+-}<yoffset>\n"
         "                   positive xoffset - offset from right display edge\n"
         "                   negative xoffset - offset from left display edge\n"
         "                   positive yoffset - offset from top display edge\n"
         "                   negative yoffset - offset from bottom display edge\n\n");
  printf("-ambientlight level  Indicates how the dark side of the globe appears:\n"
         "                     0 means totally black, 100 means totally bright (= no\n"
         "                     difference between day and night side).\n\n");
  printf("-nice priority Run the xglobe process with the given priority (see nice(1) and\n"
         "               setpriority(2) manual pages).\n\n");
  printf("-nightmap      Use default world map for the night side. The switch\n"
         "               -ambientlight will be ignored.\n"
         "-nonightmap    Disable use of night map.\n\n");
  printf("-mapfile file  Use another than the default world map. Supported image formats:\n"
#ifdef WITH_QIMAGEIO         
         "               JPG, PNG, BMP, GIF, XBM, XPM and PNM\n\n");
#else
         "               BMP, GIF, XBM, XPM and PNM\n\n");
#endif
  printf("-nightmapfile file  Same as -mapfile, but for the night map.\n\n");
  printf("-kde           Use this option when the globe disappears after switching\n"
         "               virtual screens. This is needed when using KDE.\n\n");
  printf("-dump          Saves the rendered image to \"xglobe-dump.bmp\" in the current\n"
         "               directory instead of displaying it on screen.\n\n");
  printf("-help          Display this help text.\n\n");  
}

/* ------------------------------------------------------------------------*/

void EarthApplication::init(void)
{
  QString std_marker_filename(XGLOBE_LIB_DIR);

  std_marker_filename += "/xglobe-markers";

  r = new Renderer(use_kde ? dwidget->size() : desktop()->size(),
                   (argc_map != -1) ? argv()[argc_map] : (const char *)NULL);

  if(!r)
  {
    fprintf(stderr, "Not enough memory!\n");
    exit(1);
  }

  /* initialize the Renderer */
  if(with_nightmap)
    r->loadNightMap( (argc_nightmap != -1) ? argv()[argc_nightmap] :
                                             (const char *)NULL);
  
  r->setViewPos(view_lat, view_long);
  r->setZoom(zoom);
  r->setAmbientLight(ambient_light);
  if(fov != -1.)
    r->setFov(fov);

  if(builtin_markers)
  {
    if(!marker_list.loadMarkerFile((const char *)std_marker_filename))
      ::exit(1);
  }
  r->setMarkerList(&marker_list);
  r->showMarkers(show_markers);
  r->setLabelPos(label_x, label_y);
  r->showLabel(show_label);

  timer = new QTimer(this);
  ASSERT(timer != NULL);
  
  connect(timer, SIGNAL(timeout()), this, SLOT(recalc()));
  QTimer::singleShot(1, this, SLOT(recalc())); // this will start rendering the 
  timer->start(delay*1000);                    // 1. image immediately 
}

/* ------------------------------------------------------------------------*/

void EarthApplication::recalc(void)
{
//  beep();
  if(firstTime)
  {
    firstTime = FALSE;
    time_to_render = time(NULL);      // first image with current time
    processEvents();
    r->setTime(time_to_render);
    switch(p_type)
    {
      case SUNREL:
        r->setViewPos(r->getSunLat()+view_lat, r->getSunLong()+view_long);
        break;

      case RANDOM:
        randomPosition();
        r->setViewPos(view_lat, view_long);
        break;

      default:
        break;
    }
    r->renderFrame();
  }

  if(do_the_dump)
  {
    QImage *i = r->getImage();
    i->save("xglobe-dump.bmp", "BMP");
    ::exit(0);
  }

  if(use_kde)
  {
    dwidget->updateDisplay(r->getImage());
    processEvents();                            // we want the image to be
  }                                             // displayed immediately
  else
  {
    QPixmap pm;
    pm.convertFromImage(*(r->getImage()));
    desktop()->setBackgroundPixmap(pm);
  }
  
  
  time_to_render = time(NULL)+delay;
  r->setTime(time_to_render);
  switch(p_type)
  {
    case SUNREL:
      r->setViewPos(r->getSunLat()+view_lat, r->getSunLong()+view_long);
      break;

    case RANDOM:
      randomPosition();
      r->setViewPos(view_lat, view_long);
      break;

    default:
      break;
  }
  r->renderFrame();
}

/* ------------------------------------------------------------------------*/

#include "earthapp.moc"
