/*
 * markerlist.cpp
 *
 * $Id: markerlist.cpp,v 0.1 1998/12/10 20:00:17 thorsten Exp thorsten $
 *
 * This file is part of XGlobe. See README for details.
 *
 * Copyright (C) 1998 Thorsten Scheuermann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as published by
 * the Free Software Foundation.
 *
 * $Log: markerlist.cpp,v $
 * Revision 0.1  1998/12/10 20:00:17  thorsten
 * initial revision
 *
 */

#include "markerlist.h"
#include <qfile.h>
#include <qtextstream.h>

/* ------------------------------------------------------------------------ */

Location::Location(double lon, double lat, const char *name)
{
  this->lon = lon;
  this->lat = lat;
  this->name = name;
}

/* ------------------------------------------------------------------------ */

QString Location::getName(void)
{
  return name;
}

/* ------------------------------------------------------------------------ */

double Location::getLongitude(void)
{
  return lon;
}

/* ------------------------------------------------------------------------ */

double Location::getLatitude(void)
{
  return lat;
}

/* ------------------------------------------------------------------------ */

/*
 * Loads a marker definition file.
 * This file must have the following format:
 *
 * <longitude> <latitude> <name of location>
 * where <longitude> and <latitude> are numbers specifying the geographical
 * position and <name of location> is the place's name in quotes,
 * e.g. "Atlantis".
 * Anything after these three tokens on the same line is ignored.
 * A line with a leading '#' is ignored an can be used for comments.
 */

bool MarkerList::loadMarkerFile(const char *filename)
{
  QFile f(filename);

  if(!f.open(IO_ReadOnly))
  {
    fprintf(stderr, "Error: Cant open marker file \"%s\"\n", filename);
    return FALSE;
  }
  
  QTextStream t(&f);
  QString     line;
  int         linenum = 0;
  int         numloc = 0;
  double      lon, lat;
  QString     name;
  int         pos1;
  int         pos2;

  while(!t.eof())
  {
    line = t.readLine();
    linenum++;
    line = line.simplifyWhiteSpace();

    if(line.isEmpty())                  // skip empty lines
      continue;
    
    if(((const char *)line)[0] == '#')  // if it's a comment, skip to next line
      continue;

    // read longitude
    pos1 = line.find(' ');
    if(pos1 == -1)
    {
      fprintf(stderr, "Syntax error in marker file \"%s\", line %d.\n",
              filename, linenum);
      return FALSE;
    }
    lat = line.left(pos1).toDouble();

    // read latitude
    pos2 = line.find(' ', pos1+1);
    if(pos2 == -1)
    {
      fprintf(stderr, "Syntax error in marker file \"%s\", line %d.\n",
              filename, linenum);
      return FALSE;
    }
    lon = line.mid(pos1+1, pos2-pos1-1).toDouble();

    // read name
    pos1 = line.find('\"', pos2);
    if(pos1 == -1)
    {
      fprintf(stderr, "Syntax error in marker file \"%s\", line %d.\n",
              filename, linenum);
      return FALSE;
    }
    pos2 = line.find('\"', pos1+1);
    if(pos2 == -1)
    {
      fprintf(stderr, "Syntax error in marker file \"%s\", line %d.\n",
              filename, linenum);
      return FALSE;
    }
    name = line.mid(pos1+1, pos2-pos1-1);

    // add Location to list
    append(new Location(lon, lat, name));
    numloc++;
    // printf("%.2f %.2f %s\n", lon, lat, (const char *)name);
  }

  //  printf("Read %d location definitions from file \"%s\"\n",
  //          numloc, filename);
  f.close();

  return TRUE;
}

/* ------------------------------------------------------------------------ */
