/*
 * renderer.cpp
 *
 * $Id: renderer.cpp,v 0.1 1998/12/10 20:08:57 thorsten Exp thorsten $
 *
 * This file is part of XGlobe. See README for details.
 *
 * Copyright (C) 1998 Thorsten Scheuermann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as published by
 * the Free Software Foundation.
 *
 * $Log: renderer.cpp,v $
 * Revision 0.1  1998/12/10 20:08:57  thorsten
 * initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <qapplication.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qdatetime.h>
#include <qfile.h>
#include <string.h>
#include "config.h"
#include "renderer.h"
#include "sunpos.h"

#include "marker.xpm"

#define MAX(x, y) ((x) < (y) ? (y) : (x))
#define MIN(x, y) ((x) > (y) ? (y) : (x))

/* ------------------------------------------------------------------------*/

Renderer::Renderer(const QSize &size, const char *mapfile)
{
  markerlist = NULL;
  show_markers = FALSE;
  use_nightmap = FALSE;
  map = NULL;
  mapnight = NULL;
  
  if(!mapfile)
  {
    mapFileName = XGLOBE_LIB_DIR;
    mapFileName += "/map.bmp";
  }
  else
  {
    if(QFile::exists(mapfile))
      mapFileName = mapfile;
    else
    {
      mapFileName = XGLOBE_LIB_DIR;
      mapFileName += "/";
      mapFileName += mapfile;
      if(!QFile::exists(mapFileName))
      {
        fprintf(stderr, "File \"%s\" not found!\n", (const char *)mapfile);
        ::exit(1);
      }
    }
  }

  renderedImage = new QImage(size, 32);
  if(!renderedImage)
  {
    fprintf(stderr, "Not enough memory for offscreen image buffer!\n");
    ::exit(1);
  }

  map = loadImage((const char *)mapFileName);
  
//  fprintf(stderr, "Map size: %dx%d\n", map->width(), map->height());
 
  this->radius = 1000.;
  this->view_long = 0.;
  this->view_lat = 0.;
  this->sun_long = 0.;
  this->sun_lat = 0.;
  this->fov = 0.5*PI/180.;
  this->zoom = 0.9;
  this->ambientLightLevel = 0.15;
  this->markerpixmap = new QPixmap((const char **)marker_xpm);
  ASSERT(markerpixmap != NULL);

  this->show_label = TRUE;

  calcDistance();
}

/* ------------------------------------------------------------------------*/

QImage *Renderer::loadImage(const QString &name)
{
  QImage *m;

  m = new QImage();
  if(!m)
  {
    fprintf(stderr, "Not enough memory for map!\n");
    ::exit(1);
  }

  if(!m->load(name))
  {
    fprintf(stderr, "Error while opening map \"%s\"!\n",
            (const char *)name);
    ::exit(1);
  }

  if(m->depth() != 32)
    *m = m->convertDepth(32);
 
  return m;
}

/* ------------------------------------------------------------------------*/

int Renderer::loadNightMap(const char *nmapfile)
{
  if(use_nightmap)                // we already have a night map!
    return 1;
  
  if(nmapfile)
  {
    if(QFile::exists(nmapfile))
      mapNightFileName = nmapfile;
    else
    {
      mapNightFileName = XGLOBE_LIB_DIR;
      mapNightFileName += "/";
      mapNightFileName += nmapfile;
      if(!QFile::exists(mapNightFileName))
      {
        fprintf(stderr, "File \"%s\" not found!\n", (const char *)nmapfile);
        ::exit(1);
      }
    }
  }
  else
  {
    mapNightFileName = XGLOBE_LIB_DIR;
    mapNightFileName += "/mapnight.bmp";
  }
  
  mapnight = loadImage(mapNightFileName);

  if( (mapnight->width() != map->width()) ||
      (mapnight->height() != map->height()) )
  {
    fprintf(stderr, "Day and night maps must be of equal size!\n");
    ::exit(1);
  }
  
  use_nightmap = TRUE;
  return 1;
}

/* ------------------------------------------------------------------------*/

Renderer::~Renderer()
{
  if(renderedImage)
    delete renderedImage;
  if(map)
    delete map;

  if(mapnight)
    delete mapnight;

  delete markerpixmap;
}

/* ------------------------------------------------------------------------*/

void Renderer::setViewPos(double lat, double lon)
{
  while(lat >= 360.)
    lat -= 360.;
  while(lat <= -360.)
    lat += 360.;
  if(lat > 90.)
  {
    lat = 90. - (lat - 90.);
    lon += 180.;
  }
  if(lat < -90.)
  {
    lat = -90. + (-lat - 90.);
    lon += 180.;
  }
  
  while(lon >= 360.)
    lon -= 360.;
  while(lon <= -360.)
    lon += 360.;
  if(lon > 180.)
    lon = -180. + (lon - 180.);
  if(lon < -180.)
    lon = 180. + (lon + 180.);
  
  view_lat = lat*PI/180.;
  view_long = lon*PI/180.;
}

/* ------------------------------------------------------------------------*/

double Renderer::getViewLat(void)
{
  return view_lat*180./PI;
}

/* ------------------------------------------------------------------------*/

double Renderer::getViewLong(void)
{
  return view_long*180./PI;
}

/* ------------------------------------------------------------------------*/

double Renderer::getSunLat(void)
{
  return sun_lat*180./PI;
}

/* ------------------------------------------------------------------------*/

double Renderer::getSunLong(void)
{
  return sun_long*180./PI;
}

/* ------------------------------------------------------------------------*/

void Renderer::setZoom(double z)
{
  zoom = z;
  calcDistance();
}

/* ------------------------------------------------------------------------*/

double Renderer::getZoom(void)
{
  return zoom;
}

/* ------------------------------------------------------------------------*/

void Renderer::setMarkerList(MarkerList *l)
{
  ASSERT(l != NULL);
  markerlist = l;
}

/* ------------------------------------------------------------------------*/

void Renderer::showMarkers(bool show)
{
  show_markers = show;
}

/* ------------------------------------------------------------------------*/

void Renderer::showLabel(bool show)
{
  show_label = show;
}

/* ------------------------------------------------------------------------*/

void Renderer::setLabelPos(int x, int y)
{
  label_x = x;
  label_y = y;
}

/* ------------------------------------------------------------------------*/

void Renderer::setAmbientLight(double ambient_light)
{
  ambientLightLevel = ambient_light;
}

/* ------------------------------------------------------------------------*/

void Renderer::setFov(double fov)
{
  this->fov = fov*PI/180.;
  calcDistance();
}

/* ------------------------------------------------------------------------*/

void Renderer::setTime(time_t t)
{
  time_to_render = t;
  calcLightVector();  // calc. current sun position
}

/* ------------------------------------------------------------------------*/

time_t Renderer::getTime(void)
{
  return time_to_render;
}

/* ------------------------------------------------------------------------*/

void Renderer::calcDistance(void)
{
  double x;
  double tan_a;
  
  // distance of camera to projection plane
  proj_dist = MIN(renderedImage->width(), renderedImage->height())/tan(fov);

  x = zoom*MIN(renderedImage->width(), renderedImage->height())/2.;
  tan_a = x/proj_dist;
  // distance of camera camera to center of earth ( = coordinate origin)
  center_dist = radius/sin(atan(tan_a));
}

/* ------------------------------------------------------------------------*/

void Renderer::renderFrame(void)
{
  double dir_x, dir_y, dir_z;      // direction of cast ray
  double hit_x, hit_y, hit_z;      // hit position on earth surface
  double hit2_x, hit2_y, hit2_z;   // mirrored hit position on earth surface
  double sp_x, sp_y, sp_z;         // intersection point of globe and ray
  double a, b, c;                  // coeff. of quardatic equation
  double radikand;
  double wurzel;
  double r;                        // r'
  double radiusq = radius*radius;
  double s1, s2, s;                // distance between intersections and
                                   // camera position
  double longitude, latitude;      // coordinates of hit position
  double light_angle;              // cosine of angle between sunlight and
                                   // surface normal
  int startx, endx;                // the region to be painted
  int starty, endy;

  unsigned int *p;                 // pointer to current pixel
  unsigned int *q;

  int half_width = renderedImage->width()/2 + renderedImage->width()%2;
  
  // clear image
  for(int i=0; i < renderedImage->height(); i++)
  {
    p = (unsigned int *)renderedImage->scanLine(i);
    memset(p, 0, renderedImage->bytesPerLine());
  }
  
  // rotation matrix
  double m11 = cos(view_long);
  double m12 = -sin(view_lat)*sin(view_long);
  double m13 = cos(view_lat)*sin(view_long);
//double m21 = 0.;
  double m22 = cos(view_lat);
  double m23 = sin(view_lat);
  double m31 = -sin(view_long);
  double m32 = -sin(view_lat)*cos(view_long);
  double m33 = cos(view_lat)*cos(view_long);
  
  dir_z = -proj_dist;

  // indifferent coeff.
  c = center_dist*center_dist - radiusq;

  // calc. radius of projected sphere
  b = 2*center_dist*dir_z;
  radius_proj = (int)sqrt(b*b/(4*c) - dir_z*dir_z);
  
  startx = renderedImage->width()/2 - radius_proj -1;
  startx = (startx < 0) ? 0 : startx;
  endx = renderedImage->width() - startx-1;
  starty = renderedImage->height()/2 - radius_proj -1;
  starty = (starty < 0) ? 0 : starty;
  endy = renderedImage->height() - starty-1;

  for(int py=starty; py<=endy; py++)
  {
    // handle any paint events waiting in the queue
    qApp->processEvents();
    
    startx = renderedImage->width()/2 -
             (int)sqrt(radius_proj*radius_proj -
                       (py-renderedImage->height()/2) *
                       (py-renderedImage->height()/2));
    
    startx = (startx < 0) ? 0 : startx;
    endx = renderedImage->width() - startx-1;

    // calculate offset into image data
    p = (unsigned int *)renderedImage->scanLine(py)+startx;
    q = (unsigned int *)renderedImage->scanLine(py)+endx;
    
//  for(int px=startx; px<=endx; px++)
    for(int px=startx; px<half_width; px++)
    {
      dir_x =  px - renderedImage->width()/2;
      dir_y = -py + renderedImage->height()/2;

      a = dir_x*dir_x + dir_y*dir_y + dir_z*dir_z;
      b = 2*center_dist*dir_z;
      // c constant, see above

      radikand = b*b-4*a*c;			// what's under the sq.root when solving the
                                // quadratic equation
      if(radikand >= 0.)				// solution exists <=> intersection
      {
        wurzel = sqrt(radikand);
        s1 = (-b + wurzel)/(2.*a);
        s2 = (-b - wurzel)/(2.*a);
        s = (s1<s2) ? s1 : s2;    // smaller solution belongs to nearer 
                                  // intersection
        sp_x = s * dir_x;         // sp = camera pos + s*dir
        sp_y = s * dir_y;
        sp_z = center_dist + s * dir_z;

        hit_x =  m11*sp_x + m12*sp_y + m13*sp_z;
        hit_y =             m22*sp_y + m23*sp_z;
        hit_z =  m31*sp_x + m32*sp_y + m33*sp_z;

        hit2_x = -m11*sp_x + m12*sp_y + m13*sp_z;
        hit2_y =             m22*sp_y + m23*sp_z;
        hit2_z = -m31*sp_x + m32*sp_y + m33*sp_z;

        
        longitude = atan(hit_x/hit_z);
        if(hit_z < 0.)
          longitude = PI + longitude;
        
        r = (double)sqrt(hit_x*hit_x + hit_z*hit_z);
        latitude = atan(-hit_y/r);
      
        light_angle = (light_x*hit_x + light_y*hit_y + light_z*hit_z)/radius;

        // Set pixel in image
        *p++ = getMapColorLinear(longitude, latitude, light_angle);
        
        // mirror the left half-circle of the globe: we need a new position
        // and have to recalc. the light intensity

        // this is bullshit...
        light_angle = (light_x*hit2_x + light_y*hit2_y + light_z*hit2_z)/
                      radius;
        *q-- = getMapColorLinear(2*view_long - longitude,
                                 latitude, light_angle);
			}
      else
      {
        p++;
        q--;
      }

		}
	}

  if(show_markers)
    drawMarkers();

  if(show_label)
    drawLabel();
}

/* ------------------------------------------------------------------------*/

void Renderer::calcLightVector(void)
{
  SunPos::GetSunPos(time_to_render, &sun_lat, &sun_long);

  light_x = cos(sun_lat)*sin(sun_long);
  light_y = sin(sun_lat);
  light_z = cos(sun_lat)*cos(sun_long);
}

/* ------------------------------------------------------------------------*/

unsigned int Renderer::getMapColorLinear(double longitude, double latitude,
                                         double angle)
{
	unsigned int *p1;
	unsigned int *p2;

  // for night map
	unsigned int *p3 = NULL;
	unsigned int *p4 = NULL;
  
	if(longitude < 0.)
		longitude += 2*PI;

	latitude += PI/2;

	longitude += PI;
	if(longitude >= 2*PI)
		longitude -= 2*PI;

	int x11 = (int)(longitude*map->width()/(2*PI));
	int y11 = (int)(latitude*map->height()/PI);
	int x12 = (x11+1) % map->width();
	int y12 = y11;
	int x21 = x11;
	int y21 = (y11+1) % map->height();
	int x22 = x12;
	int y22 = y21;
	double dx = (longitude*map->width()/(2*PI))-x11;
	double dy = (latitude*map->height()/PI)-y11;
  x11 %= map->width();
  y11 %= map->height();

	double r1,g1,b1, r2,g2,b2, r,g,b;

  // for night map
  double r3 = 0.;
  double g3 = 0.;
  double b3 = 0.;
  double r4 = 0.;
  double g4 = 0.;
  double b4 = 0.;

  // offset into map pixel data
	p1 = (unsigned int *)map->scanLine(y11) + x11;
	p2 = (unsigned int *)map->scanLine(y12) + x12;

  if(use_nightmap)
  {
    p3 = (unsigned int *)mapnight->scanLine(y11) + x11;
    p4 = (unsigned int *)mapnight->scanLine(y12) + x12;
  }
  
	int sb = qBlue(*p1);
	int sg = qGreen(*p1);
	int sr = qRed(*p1);
	int tb = qBlue(*p2);
	int tg = qGreen(*p2);
	int tr = qRed(*p2);
	r1 = (tr-sr)*dx+sr;
	g1 = (tg-sg)*dx+sg;
	b1 = (tb-sb)*dx+sb;
  if(use_nightmap)
  {
    sb = qBlue(*p3);
    sg = qGreen(*p3);
    sr = qRed(*p3);
    tb = qBlue(*p4);
    tg = qGreen(*p4);
    tr = qRed(*p4);
    r3 = (tr-sr)*dx+sr;
    g3 = (tg-sg)*dx+sg;
    b3 = (tb-sb)*dx+sb;
  }

	p1 = (unsigned int *)map->scanLine(y21) + x21;
	p2 = (unsigned int *)map->scanLine(y22) + x22;
  if(use_nightmap)
  {
    p3 = (unsigned int *)mapnight->scanLine(y21) + x21;
    p4 = (unsigned int *)mapnight->scanLine(y22) + x22;
  }
  
	sb = qBlue(*p1);
	sg = qGreen(*p1);
	sr = qRed(*p1);
	tb = qBlue(*p2);
	tg = qGreen(*p2);
	tr = qRed(*p2);
	r2 = (tr-sr)*dx+sr;
	g2 = (tg-sg)*dx+sg;
	b2 = (tb-sb)*dx+sb;
  if(use_nightmap)
  {
    sb = qBlue(*p3);
    sg = qGreen(*p3);
    sr = qRed(*p3);
    tb = qBlue(*p4);
    tg = qGreen(*p4);
    tr = qRed(*p4);
    r4 = (tr-sr)*dx+sr;
    g4 = (tg-sg)*dx+sg;
    b4 = (tb-sb)*dx+sb;
  }
  
	r = (r2-r1)*dy+r1;
	g = (g2-g1)*dy+g1;
	b = (b2-b1)*dy+b1;
  if(use_nightmap)
  {
    r1 = (r4-r3)*dy+r3;
    g1 = (g4-g3)*dy+g3;
    b1 = (b4-b3)*dy+b3;
  }
  
  if(use_nightmap)
  {
    if(angle > 0)
    {
      r = r1*(1.0-angle) + r*angle;
      g = g1*(1.0-angle) + g*angle;
      b = b1*(1.0-angle) + b*angle;
    }
    else
    {
      r = r1;
      g = g1;
      b = b1;
    }
  }
  else
  {
    if(angle > 0)
    {
      r = r*ambientLightLevel + r*angle*(1.-ambientLightLevel);
      g = g*ambientLightLevel + g*angle*(1.-ambientLightLevel);
      b = b*ambientLightLevel + b*angle*(1.-ambientLightLevel);
//		b *= angle;
    }
    else
    {
      r *= ambientLightLevel;
      g *= ambientLightLevel;
      b *= ambientLightLevel;
    }
  }
  
  return qRgb((int)r,(int)g,(int)b);
}

/* ------------------------------------------------------------------------*/

void Renderer::drawMarkers(void)
{
  Location *l;
  double lon, lat;
  double s_x, s_y, s_z;
  double loc_x, loc_y, loc_z;
  int screen_x, screen_y;
  double m11, m12, m13,
         m21, m22, m23,
         m31, m32, m33;
  double visible_angle;

  int i, num;
  Location **visible_locations;

  visible_locations = new Location*[markerlist->count()];
  ASSERT(visible_locations != NULL);

  // Matrix M of renderFrame, but transposed
  m11 = cos(view_long);
  m12 = 0.;
  m13 = -sin(view_long);
  m21 = -sin(view_lat)*sin(view_long);
  m22 = cos(view_lat);
  m23 = -sin(view_lat)*cos(view_long);
  m31 = cos(view_lat)*sin(view_long);
  m32 = sin(view_lat);
  m33 = cos(view_lat)*cos(view_long);

  visible_angle = radius/center_dist;
  
  for(i=0, l = markerlist->first(); l != NULL; l = markerlist->next())
  {
    lon = l->getLongitude()*PI/180.;
    lat = l->getLatitude()*PI/180.;

    s_x = radius*cos(lat)*sin(lon);
    s_y = radius*sin(lat);
    s_z = radius*cos(lat)*cos(lon);
    loc_x =  m11*s_x + m12*s_y + m13*s_z;
    loc_y =  m21*s_x + m22*s_y + m23*s_z;
    loc_z =  m31*s_x + m32*s_y + m33*s_z;
    
    l->cos_angle = loc_z/radius;
    
    if(l->cos_angle < visible_angle)
      // location lies on the other side
      continue;

    loc_y = -loc_y;
    loc_z = center_dist - loc_z;
    screen_x = (int)(loc_x*proj_dist/loc_z);
    screen_y = (int)(loc_y*proj_dist/loc_z);
    screen_x += renderedImage->width()/2;
    screen_y += renderedImage->height()/2;

    if((screen_x < 0) || (screen_x >= renderedImage->width()))
      // location out of bounds
      continue;
    if((screen_y <0) || (screen_y >= renderedImage->height()))
      continue;

    l->x = screen_x;
    l->y = screen_y;

    visible_locations[i] = l;
    i++;
  }

  num = i;

  // sort the markers according to depth
  qsort(visible_locations, num, sizeof(Location *),
        Renderer::compareLocations);

  for(i=0; i<num; i++)
    paintMarker(visible_locations[i]->x,
                visible_locations[i]->y, visible_locations[i]);

  delete[] visible_locations;
}

/* ------------------------------------------------------------------------*/

int Renderer::compareLocations(Location **l1, Location **l2)
{
  double c1, c2;

  c1 = (*l1)->cos_angle;
  c2 = (*l2)->cos_angle;
  
  if(c1 > c2)
    return 1;
  if(c1 < c2)
    return -1;
  return 0;
}

/* ------------------------------------------------------------------------*/

QImage *Renderer::getImage(void)
{
  QImage *clonedImage = NULL;

  clonedImage = new QImage(*renderedImage);
  ASSERT(clonedImage != NULL);
  return clonedImage;
}

/* ------------------------------------------------------------------------*/

void Renderer::paintMarker(int x, int y, Location *l)
{
  QPainter p;
  int wx, wy;
  unsigned int *src, *dest;
  unsigned int transparentcolor = qRgb(255, 255, 0);  

  QFontMetrics fm(QFont("helvetica", 12, QFont::Bold));
  
  QRect br = fm.boundingRect(l->getName());
  QPixmap pm(markerpixmap->width()+2 + br.width()+2,
             MAX(markerpixmap->height(), br.height()+2));
  
  p.begin(&pm);
  p.setFont(QFont("helvetica", 12, QFont::Bold));
  p.fillRect(0,0,pm.width(), pm.height(), QColor(255, 255, 0));

  // draw a black seam around the text
  p.setPen(black);
  wx = -br.x()+markerpixmap->width()+2;
  wy = -br.y();
  p.drawText(wx, wy+1, l->getName());
  p.drawText(wx+1, wy, l->getName());
  p.drawText(wx+1, wy+2, l->getName());
  p.drawText(wx+2, wy+1, l->getName());


  p.setPen(red);
  p.drawText(wx+1, wy+1, l->getName());
  
  p.drawPixmap(0, (pm.height() - markerpixmap->height())/2,
               *markerpixmap);
  
  p.end();
  
  QImage labelimage = pm.convertToImage();
  if(labelimage.depth() != 32)
    labelimage = labelimage.convertDepth(32);

  x -= markerpixmap->width()/2;
  y -= labelimage.height()/2;

  QRect screenrect(0, 0, renderedImage->width(), renderedImage->height());
  QRect labelrect(x, y, pm.width(), pm.height());
  QRect visiblerect = screenrect.intersect(labelrect);

  if(visiblerect.isEmpty())
    // the label is not visible
    return;
  
  for(wy=0 ; wy<visiblerect.height(); wy++)
  {
    dest = (unsigned int *)renderedImage->scanLine(visiblerect.y()+wy) +
           visiblerect.x();
    src = (unsigned int *)labelimage.scanLine(visiblerect.y()-y+wy) +
          (visiblerect.x()-x);
    
    for(wx=0 ; wx<visiblerect.width(); wx++)
    {      
      if(*src != transparentcolor)
        *dest++ = *src++;
      else
      {
        src++;
        dest++;
      }      
    }
  }  
}

/* ------------------------------------------------------------------------*/

void Renderer::drawLabel(void)
{
  
  
  QDateTime    dt;
  QString      labelstring;
  QPainter     p;
  unsigned int transparentcolor = qRgb(255, 255, 0);
  unsigned int pixel;
  int          x, y;
  int          wx, wy;
  unsigned int *src, *dest;
  double vlon, vlat;
  double slon, slat;
  struct tm    *tm;
  
  dt.setTime_t(time_to_render);
  tm = localtime(&time_to_render);
  
  vlon = view_long*180./PI;
  vlat = view_lat*180./PI;
  slon = sun_long*180./PI;
  slat = sun_lat*180./PI;
  
  labelstring.sprintf("%s, %s %d. %d, %d:%02d %s\n"
                      "View pos %2.2f %c %2.2f %c\n"
                      "Sun pos %2.2f %c %2.2f %c",
                      dt.date().dayName(dt.date().dayOfWeek()),
                      dt.date().monthName(dt.date().month()),
                      dt.date().day(), dt.date().year(),
                      dt.time().hour(), dt.time().minute(),
                      tzname[tm->tm_isdst],
                      fabs(vlat), (vlat < 0.) ? 'S' : 'N',
                      fabs(vlon), (vlon < 0.) ? 'W' : 'E',
                      fabs(slat), (slat < 0.) ? 'S' : 'N',
                      fabs(slon), (slon < 0.) ? 'W' : 'E');

  QFontMetrics fm(QFont("helvetica", 12, QFont::Bold));
  
  QRect br = fm.boundingRect(0, 0, 0, 0, AlignLeft|AlignTop, labelstring);
  QPixmap pm(br.width()+10, br.height()+10);
  
  p.begin(&pm);
  p.setFont(QFont("helvetica", 12, QFont::Bold));
  p.fillRect(0,0,pm.width(), pm.height(), QColor(255, 255, 0));
  
  p.setPen(white);
  p.drawText(5, 5, br.width(), br.height(), AlignLeft|AlignTop, labelstring);
  p.end();
  
  QImage labelimage = pm.convertToImage();
  if(labelimage.depth() != 32)
    labelimage = labelimage.convertDepth(32);

  if(label_x > 0)
    x = label_x;
  else
    x = renderedImage->width()-labelimage.width()+label_x;
  if(label_y > 0)
    y = label_y;
  else
    y = renderedImage->height()-labelimage.height()+label_y;
  
  for(wy=0 ; wy<labelimage.height(); wy++)
  {
    dest = (unsigned int *)renderedImage->scanLine(y+wy) + x;
    src = (unsigned int *)labelimage.scanLine(wy);
    
    for(wx=0 ; wx<labelimage.width(); wx++)
    {      
      if(*src != transparentcolor)
        *dest++ = *src++;
      else
      {
        pixel = *dest;
        src++;
        *dest++ = qRgb(qRed(pixel)/2, qGreen(pixel)/2, qBlue(pixel)/2);
      }      
    }
  }
} 

/* ------------------------------------------------------------------------*/
